@@ -9,7 +9,7 @@ from .datasets import get_data
 from .distribution import get_workers
 from .utils import create_logging
 
-def load_config(config_file=None,debug=0,operation=None,**kwargs):
+def load_config(config_file=None,debug=0,verbose=False,operation=None,no_submit=False,**kwargs):
     """
     Load configuration file and embed relevant parameters and data in dictionary.
     """
@@ -40,8 +40,10 @@ def load_config(config_file=None,debug=0,operation=None,**kwargs):
         'backend':None,
         'split':'trial',
         'log_dir':'logs',
+        'no_submit':no_submit,
+        'system':'cori',
     }
-    if 'SLURM_NTASKS' in os.environ.keys() or operation=='slurm':
+    if 'SLURM_NTASKS' in os.environ.keys() or operation=='sbatch':
         assert 'dist' in config.keys(), 'Distribution settings not defined. Abort.'
         config['dist'] = {**default,**config['dist']}
     else:
@@ -50,7 +52,7 @@ def load_config(config_file=None,debug=0,operation=None,**kwargs):
     if operation in ['evaluation','surrogate']:
         config['dist'] = {**config['dist'],**get_workers(**config)}
         config['dist'].pop('operation', None)
-        create_logging(operation,**config['dist'])
+        create_logging(operation,verbose,**config['dist'])
         # Extract training and test datasets
         if 'data' in config.keys():
             # Merging needed to avoid duplicate (e.g. verbose)
