@@ -1,6 +1,7 @@
 # System
 import os
 import time
+import copy
 import random
 import logging
 import importlib
@@ -116,27 +117,24 @@ def run_uq(trials, hyperprms, data, output, loss, log_dir, dl_type, trial, **kwa
       UQ information provided by the user in the configuration file (
       :ref:`see here <uq>` for more details).
     """
+    logging.info('-'*40)
+    logging.info('UNCERTAINTY QUANTIFICATION {:>6} MODELS'.format(len(trials)))
+    logging.info('-'*40)
     device = torch.device('cpu')
     # Architecture initialization
     module = importlib.import_module('.' + dl_type.lower(), 'hyppo.dnnmodels.pytorch')
     net = module.get_model(data=data['train'], prms=hyperprms, **kwargs).to(device)
     trained_models = []
     for i in trials:
-        model = os.path.join(log_dir,'checkpoints',output,'%02i.pth.tar' % (i+1))
-        trained_models.append(load_model(model,net,device))
+        path = os.path.join(log_dir,'checkpoints',output,'%02i.pth.tar' % (i+1))
+        trained_models.append(load_model(path,net,device))
     dropout_rate = hyperprms.get('dropout',0.05)[0]
     if dropout_rate == 0:
         logging.warning('Running uncertainty quantification with dropout rate 0 is not optimal.')
         logging.warning('Dropout rate set to 5% for Monte Carlo dropout UQ analysis as a default.')
     criterion = hyperprms['loss'](**hyperprms['loss_args'])
-    pred_mean, total_var, confidence_interval = uq_quantifier(output, log_dir, data, trained_models, criterion, **kwargs)
-    logging.info('-'*40)
-    logging.info('OUTER OBJECTIVE FUNCTION {:>8} TRIALS'.format(len(loss)))
-    logging.info('-'*40)
-    logging.info('\tOuter Loss           : {:>9.5f}'.format(confidence_interval[0]))
-    logging.info('\tLower Bound of CI    : {:>9.5f}'.format(confidence_interval[1]))
-    logging.info('\tUpper Bound of CI    : {:>9.5f}'.format(confidence_interval[2]))
-    return confidence_interval[0] #{'pred_mean':pred_mean, 'pred_var': total_var,'confidence_int':confidence_interval}
+    pred_mean_obj_val, pred_std = uq_quantifier(output, log_dir, data, trained_models, criterion, **kwargs)
+    return pred_mean_obj_val, pred_std
   
 def load_model(fname,model,device='cpu'):
     """
@@ -157,6 +155,7 @@ def load_model(fname,model,device='cpu'):
     model : :py:class:`torch.nn.Module`
       Up-to-date neural network model
     """
+    model = copy.deepcopy(model)
     state_dict = torch.load(fname, map_location=device)
     new_state_dict = OrderedDict()
     for key, value in state_dict.items():
@@ -164,61 +163,77 @@ def load_model(fname,model,device='cpu'):
             key = key[7:]
         new_state_dict[key] = value
     model.load_state_dict(new_state_dict)
+    logging.debug('Model %s with sumw %i' % (fname.split('/')[-1],sum(p.sum() for p in model.parameters())))
     return model
 
 def uq_quantifier(output, log_dir, data, trained_models, criterion, uq_weights=[0.5,0.5],
-                  dropout_masks=30, data_noise=0, data_type=None, update=False, device=None, **kwargs):
-    data_variance = data_noise
-    weights = uq_weights
+                  dropout_masks=30, data_noise=0, update=False, device=None, **kwargs):
+    """
+    Main Uncertainty Quantification (UQ) algorithm.
+    """
     T = len(trained_models)
     alt_models=[]
-    if data_variance == 0.0:
+    if data_noise == 0.0:
         for i in range(T):
+            logging.debug('Applying Dropout Model Builder to model %i' % (i+1))
             alt_models.append(Dropout_Model_Builder(trained_models[i]))
             # fake_model = trained_models[i]
             # alt_models.append(fake_model)
     else:
         for i in range(T):
-            alt_models.append(ADF_Dropout_Model_Builder(trained_models[i],data_variance))
+            logging.debug('Applying ADF Dropout Model Builder to model %i' % (i+1))
+            alt_models.append(ADF_Dropout_Model_Builder(trained_models[i],data_noise))
     uq_means = {'trained':{}, 'Unc':{}}
     uq_vars  = {'trained':{}, 'Unc':{}}
     uq_obj_vals = []
     N = dropout_masks
-    logging.info('-'*40)
-    logging.info('UNCERTAINTY QUANTIFICATION {:>6} MODELS'.format(T))
-    logging.info('-'*40)
-    logging.info('Model | Obj. Value | Dropout Mask Time')
     for j in range(T):
-        if data_variance > 0.0:
-            model = ADF_Dropout_Model_Builder(trained_models[j],data_variance,True)
+        logging.debug('-'*40)
+        model_vals = []
+        logging.debug('UQ results on model %i/%i:' % (j+1,T))
+        if data_noise > 0.0:
+            model = ADF_Dropout_Model_Builder(trained_models[j],data_noise,True)
         else:
             model = trained_models[j]
             model.eval()
+        logging.debug('  Parameter sum: %i' % sum(p.sum() for p in model.parameters()))
+        logging.debug('  Prediction evaluation')
         trained_val, trained_pred, trained_var, y_real = pred_evaluate(data['test'], model, criterion,update,device,**kwargs)
-        uq_obj_vals.append(trained_val)
+        model_vals.append(trained_val)
         uq_means['trained'][j] = numpy.asarray(trained_pred)
         uq_vars['trained'][j]  = numpy.asarray(trained_var)
-        start_time = time.time()
         model = alt_models[j]
         model.train()
+        start_time = time.time()
         for t in range(N): 
-            val, pred, pred_var, y_real= pred_evaluate(data['test'], model, criterion, update, device, **kwargs)
-            uq_obj_vals.append(val)
+            logging.debug('  Dropout mask %i/%i' % (t+1,N))
+            val, pred, pred_var, y_real = pred_evaluate(data['test'], model, criterion, update, device, **kwargs)
+            model_vals.append(val)
             uq_means['Unc'][j*N+t] = numpy.asarray(pred)
             uq_vars['Unc'][j*N+t] = numpy.asarray(pred_var)
-        logging.info('{:>5} | {:>10.5f} | {:>17.3f} s'.format(j+1,trained_val.detach(),time.time()-start_time))
-    pred_mean = (weights[0]/T)*sum(uq_means['trained'].values())+weights[1]*sum(uq_means['Unc'].values())/(T*N)
-    data_var  = (weights[0]/T)*sum(uq_vars['trained'].values())+weights[1]*sum(uq_vars['Unc'].values())/(T*N)
+        logging.debug('-'*40)
+        logging.info('Model {:>2} | Loss {:>8.5f} | {:>10.3f} s'.format(j+1,numpy.nanmean(torch.stack(model_vals).detach().numpy()),time.time()-start_time))
+        uq_obj_vals.extend(model_vals)
+    pred_mean = (uq_weights[0]/T)*sum(uq_means['trained'].values())+uq_weights[1]*sum(uq_means['Unc'].values())/(T*N)
+    save_results(T,N,uq_weights,uq_means,uq_vars,pred_mean,y_real,output,**kwargs)
+    # Return statistics
+    pred_mean_obj_val = criterion(torch.tensor(pred_mean),y_real)
+    pred_std = numpy.std(torch.stack(uq_obj_vals).cpu().detach().numpy())
+    return pred_mean_obj_val, pred_std
+
+def save_results(T,N,uq_weights,uq_means,uq_vars,pred_mean,y_real,output,data_type=None,log_dir='logs',**kwargs):
+    data_var = (uq_weights[0]/T)*sum(uq_vars['trained'].values())+uq_weights[1]*sum(uq_vars['Unc'].values())/(T*N)
+    logging.debug('-'*40)
+    logging.debug('Average predicted mean  : %.5f' % numpy.average(pred_mean))
+    logging.debug('Average output variance : %.5f' % numpy.average(data_var))
     model_trained_var = torch.square(torch.tensor(uq_means['trained'][0]) - torch.tensor(pred_mean)) 
     model_Unc_var     = torch.square(torch.tensor(uq_means['Unc'][0]) - torch.tensor(pred_mean)) 
     for i  in range(1,T):
         model_trained_var = model_trained_var + torch.square(torch.tensor(uq_means['trained'][i]) - torch.tensor(pred_mean))
     for j in range(1,T*N):
         model_Unc_var = model_Unc_var + torch.square(torch.tensor(uq_means['Unc'][j]) - torch.tensor(pred_mean)) 
-    model_var = (weights[0]/T)*numpy.asarray(model_trained_var)+ (weights[1]/(N*T))*numpy.asarray(model_Unc_var)
+    model_var = (uq_weights[0]/T)*numpy.asarray(model_trained_var)+ (uq_weights[1]/(N*T))*numpy.asarray(model_Unc_var)
     total_var = data_var + model_var
-    pred_mean_obj_val = criterion(torch.tensor(pred_mean),y_real)
-    pred_std = numpy.std(torch.stack(uq_obj_vals).cpu().detach().numpy())
     if data_type=='im':
         path = os.path.join(log_dir,'output',output)
         os.makedirs(path,exist_ok=True)
@@ -235,23 +250,8 @@ def uq_quantifier(output, log_dir, data, trained_models, criterion, uq_weights=[
         numpy.savetxt(os.path.join(path,'pred_mean.txt'),pred_mean.squeeze(),fmt='%f')
         numpy.savetxt(os.path.join(path,'pred_var.txt'),total_var.squeeze(),fmt='%f')
 
-    #print('uq_obj_val={}'.format(numpy.asarray(uq_obj_vals)))
-    #print('pred_mean_obj_val  ={}'.format(pred_mean_obj_val))
-    #print('confidence intevals = [{},{},{}]'.format(pred_mean_obj_val-2*pred_std, pred_mean_obj_val, pred_mean_obj_val+2*pred_std))
-    #print('pred_mean = {}'.format(pred_mean.squeeze()))
-    #print('total var = {}'.format(total_var.squeeze()))
-    #print('number correct:{}'.format(numpy.count_nonzero((numpy.asarray(y_real)-numpy.argmax(numpy.asarray(uq_means['trained'][0]),axis=1))==0)))
-    #print('trained model pred={}'.format(numpy.argmax(numpy.asarray(uq_means['trained'][0]),axis=1)))
-
-    # Return statistics
-    lb = pred_mean_obj_val-2*pred_std
-    ub = pred_mean_obj_val+2*pred_std
-    confidence_interval = [pred_mean_obj_val,lb,ub]
-    
-    return pred_mean, total_var, confidence_interval
-
-############################################################################################
-# These classes construct the layers which propogate data uncertianty using ADF
+#-------------------------------------------------------------------------------------------
+# These classes construct the layers which propogate data uncertainty using ADF
 #-------------------------------------------------------------------------------------------
 class ADF_Linear_Layer(torch.nn.Module):
     def __init__(self,Linear):
@@ -331,7 +331,6 @@ def adf_layer_selector(model, lay_num, lay_name, dropout_rate):
         layer = ADF_Other_Act_Layer(lay_name)
     return layer
 
-############################################################################################
 class ADF_Dropout_Model_Builder(torch.nn.Module):
     
     def __init__(self, model, data_noise, exact = False):
@@ -378,7 +377,6 @@ class ADF_Dropout_Model_Builder(torch.nn.Module):
             out_var[i]  = out[1,:]
         return {'mean':out_mean, 'var': out_var} 
 
-############################################################################################
 class Dropout_Model_Builder(torch.nn.Module):
 
     def __init__(self, model, **kwargs):
@@ -462,7 +460,13 @@ def pred_evaluate(data, model, loss_function, update=False, device=None, **kwarg
         y_real = torch.stack(y_real)
     y_pred_mean = torch.stack(y_pred_mean)
     loss = loss_function(y_pred_mean.to(device), y_real.to(device))
-    return loss, y_pred_mean.cpu().detach().numpy(), torch.tensor(numpy.asarray(y_pred_var)), y_real
+    trained_pred = y_pred_mean.cpu().detach().numpy()
+    trained_var = torch.tensor(numpy.asarray(y_pred_var))
+    logging.debug('    trained_val : %.5f' % (loss))
+    logging.debug('    trained_pred: %.5f' % (numpy.average(trained_pred)))
+    logging.debug('    trained_var : %.5f' % (numpy.average(trained_var)))
+    logging.debug('    y_real      : %.5f' % (numpy.average(y_real)))
+    return loss, trained_pred, trained_var, y_real
 
 def time_series_uq_plotter(output, data, trained_means, pred_mean, pred_var, uq_means, plot_title):
     train_data   = data['train']
