@@ -72,23 +72,30 @@ def train_evaluation(x_sc, samples, trainer, model, dist, output='out', data={},
             continue
         losses = gather_trials(samples,output,**model,**dist)
         if 'uq_on' in uq.keys() and uq['uq_on']==True:
-            valid_trials = numpy.argwhere(~numpy.isnan(losses)).squeeze()
-            res = run_uq(trials=valid_trials,hyperprms=hyperprms,data=loaders,output=output,loss=losses,**model,**dist,**uq)
+            valid_trials = numpy.argwhere(~numpy.isnan(losses) & ~numpy.isinf(losses)).squeeze()
+            inner_loss, unc = run_uq(trials=valid_trials,hyperprms=hyperprms,data=loaders,output=output,loss=losses,**model,**dist,**uq)
+            if numpy.isnan(unc) or numpy.isinf(unc) or numpy.isnan(inner_loss):
+                outer_loss = float('nan')
+                inner_loss = float('nan')
+                unc        = float('nan')
+            else:
+                outer_loss = calculate_outer_loss(inner_loss,**model)
         else:
             if numpy.isnan(losses).all():
                 outer_loss = float('nan')
                 inner_loss = float('nan')
-                std        = float('nan')
+                unc        = float('nan')
             else:
                 outer_loss = numpy.nanmean(calculate_outer_loss(losses,**model))
                 inner_loss = numpy.nanmean(losses)
-                std        = numpy.nanstd(losses)
-            logging.info('-'*40)
-            logging.info('OUTER OBJECTIVE FUNCTION {:>8} TRIALS'.format(len(losses)))
-            logging.info('-'*40)
-            logging.info('\tOuter Loss : {:>11.5f}'.format(outer_loss))
-            logging.info('\tInner Loss : {:>11.5f}'.format(inner_loss))
-            logging.info('\tSTD        : {:>11.5f}'.format(std))
+                unc        = numpy.nanstd(losses)
+        res = outer_loss
+        logging.info('-'*40)
+        logging.info('OUTER OBJECTIVE FUNCTION {:>8} TRIALS'.format(len(losses)))
+        logging.info('-'*40)
+        logging.info('{:>14} : {:>11.5f}'.format('Outer Loss',outer_loss))
+        logging.info('{:>14} : {:>11.5f}'.format('Inner Loss',inner_loss))
+        logging.info('{:>14} : {:>11.5f}'.format('Uncertainty',unc))
     else:
         logging.info('-'*40)
         logging.info('Waiting for other processes to finish...')
@@ -103,12 +110,15 @@ def external_training(itrial, x_sc, trainer, output_dir, model, log_dir, **kwarg
     logging.info('-'*40)
     logging.info('{} {:>3} / {:<3} {:>24}'.format('TRIAL',itrial+1,model['trial'],'TRAINING'))
     logging.info('-'*40)
-    model['output_dir'] = os.path.join(log_dir,'output',output_dir)
+    if 'output_dir' not in model.keys():
+        model['output_dir'] = os.path.join(log_dir,'output',output_dir)
+    else:
+        model[model['output_dir']] = os.path.join(log_dir,'output',output_dir)
     if 'yaml_input' in model.keys():
         params = yaml_creation(x_sc,model['output_dir'],**model['yaml_input'])
     else:
-        params = {**model,**x_sc}
-    if 'test_problem' in trainer:
+        params = {**model,**x_sc,**kwargs}
+    if '.' not in trainer and 'test_problem' in trainer:
         loss = eval(trainer)(**params)
     elif type(trainer)==types.FunctionType:
         loss = trainer(**params)
@@ -121,6 +131,7 @@ def external_training(itrial, x_sc, trainer, output_dir, model, log_dir, **kwarg
     logging.info('-'*40)
     logging.info('\t Test Loss : {:>11.5f}'.format(loss))
     logging.info('\tTrain Time : {:>11.5f} s'.format(time.time()-start_time))
+    return loss
 
 def yaml_creation(x_sc,output_dir,src,config=None,**kwargs):
     # Extract input configuration file
