@@ -12,7 +12,23 @@ import pandas as pd
 import pickle5 as pickle
 from scipy import interpolate
 
-def get_convergence(search_dir,with_surrogate=False,werr=True,niters=None,nruns=None,debug=False,median=False,**kwargs):
+# Local
+from ..datasets import get_data, get_loader
+from ..hyperparams import get_hyperprms
+from ..dnnmodels import inference
+
+def single_convergence(solutions):
+    loss_surr, sdev_surr = [], []
+    for i in range(len(solutions)):
+        if len(loss_surr)==0 or solutions.iloc[i].loss<loss_surr[-1]:
+            loss_surr.append(solutions.iloc[i].loss)
+            sdev_surr.append(solutions.iloc[i].stdev)
+        else:
+            loss_surr.append(loss_surr[-1])
+            sdev_surr.append(sdev_surr[-1])
+    return numpy.array(loss_surr), numpy.array(sdev_surr)
+
+def mult_convergence(search_dir,with_surrogate=True,werr=True,niters=None,nruns=None,debug=False,median=False,**kwargs):
     dir_list = sorted(glob.glob(search_dir+'/*'))
     if any('.log' in dir_name for dir_name in dir_list):
         dir_list = [search_dir]
@@ -32,14 +48,7 @@ def get_convergence(search_dir,with_surrogate=False,werr=True,niters=None,nruns=
         if with_surrogate:
             surrogate  = extract(log_files.replace('evaluation','surrogate'),debug=debug,**kwargs).sort_values(by=['nevals']).to_numpy()[:,1:3]
             evaluation = numpy.vstack((evaluation,surrogate))
-        loss_surr, sdev_surr = [], []
-        for loss,std in evaluation:
-            if len(loss_surr)==0 or loss+std/2<loss_surr[-1]+sdev_surr[-1]/2:
-                loss_surr.append(loss)
-                sdev_surr.append(std)
-            else:
-                loss_surr.append(loss_surr[-1])
-                sdev_surr.append(sdev_surr[-1])
+        loss_surr, sdev_surr = single_convergence(evaluation)
         all_results[0].append(loss_surr)
         all_results[1].append(sdev_surr)
         n+=1
@@ -60,8 +69,59 @@ def get_convergence(search_dir,with_surrogate=False,werr=True,niters=None,nruns=
         mean_sdev = numpy.std(all_results[0],axis=0)
     return mean_loss, mean_sdev
 
-def extract(path,target_loss='inner',target_unc='STD',debug=False,**kwargs):
+def extract(path,target_loss='inner',target_unc='Uncertainty',debug=False,raw=False,**kwargs):
     hps = []
+    data = []
+    nevals = 0
+    for fname in sorted(glob.glob(path)):
+        if debug:
+            print('Extract results from',fname)
+        network_names, network_prms = [], []
+        content = numpy.loadtxt(fname,dtype=str,delimiter='\n')
+        for i,line in enumerate(content):
+            if 'CONFIGURATION:' in line:
+                cfg = ast.literal_eval(content[i+2].strip())
+                names = cfg['prms']['names']
+                mult = cfg['prms']['mult']
+            if 'Samples:' in line:
+                samples = line.split('Samples:')[-1].strip()[1:-1]
+                samples = [int(k) for k in samples.split()]
+                if raw:
+                    names = ['hp%i'%(i+1) for i in range(len(samples))]
+                else:
+                    samples = [int(samples[n])*mult[n] for n in range(len(mult))]
+                network_names, network_prms = [], []
+            if 'parameters in' in line:
+                network_names.append(line.split()[-1])
+                network_prms.append(int(line.split()[-4].replace(',','')))
+            if 'Number of parameters :' in line:
+                network_names.append('tot_prms')
+                network_prms.append(int(line.split()[-1].replace(',','')))
+            if '%s Loss' % target_loss.capitalize() in line:
+                loss = float(line.split()[-1])
+            if 'EVALUATION' in line:
+                nevals = int(line.split()[-3])
+            if 'Total of sample sets used' in line:
+                nevals = int(line.split()[-1])+1
+            if target_unc in line:
+                std = float(line.split()[-1])
+                time = float(content[i+2].split()[-2])
+                data.append([loss,std]+network_prms+samples+[nevals,time])
+                hps.append(samples)
+            if 'Lower Bound of CI' in line:
+                std = loss-float(line.split()[-1])
+                time = float(content[i+2].split()[-2])
+                data.append([loss,std]+network_prms+samples+[nevals,time])
+                hps.append(samples)
+    if len(data)!=len(numpy.unique(hps,axis=0)):
+        print('WARNING: Some hyperparameter sets were evaluated more than once in %s (%i/%i)' % (path,len(numpy.unique(hps,axis=0)),len(data)))
+    if len(data)>0:
+        data = pd.DataFrame(data=data,columns=['loss','stdev']+network_names+names+['nevals','time']).sort_values(by=['nevals'])
+    if debug:
+        print(data)
+    return data
+
+def extract_trials(path,debug=False,**kwargs):
     data = []
     nevals = 0
     for fname in sorted(glob.glob(path)):
@@ -73,30 +133,25 @@ def extract(path,target_loss='inner',target_unc='STD',debug=False,**kwargs):
             if 'Samples:' in line:
                 samples = line.split('Samples:')[-1].strip()[1:-1]
                 samples = [int(k) for k in samples.split()]
-            if 'parameters in generator' in line:
-                gen_params = int(line.split()[-4].replace(',',''))
-            if 'parameters in discriminator' in line:
-                dis_params = int(line.split()[-4].replace(',',''))
+                network_names, network_prms = [], []
+            if 'parameters in' in line:
+                network_names.append(line.split()[-1])
+                network_prms.append(int(line.split()[-4].replace(',','')))
             if 'Number of parameters :' in line:
-                params = int(line.split()[-1].replace(',',''))
-            if '%s Loss' % target_loss.capitalize() in line:
-                loss = float(line.split()[-1])
+                network_names.append('tot_prms')
+                network_prms.append(int(line.split()[-1].replace(',','')))
+            if 'TRIAL ' in line:
+                trial = int(line.split()[-4])
             if 'EVALUATION' in line:
                 nevals = int(line.split()[-3])
             if 'Total of sample sets used' in line:
                 nevals = int(line.split()[-1])+1
-            if target_unc in line:
-                std = float(line.split()[-1])
-                data.append([params,loss,std]+samples+[nevals,gen_params,dis_params])
-                hps.append(samples)
-            if 'Lower Bound of CI' in line:
-                std = loss-float(line.split()[-1])
-                data.append([params,loss,std]+samples+[nevals,gen_params,dis_params])
-                hps.append(samples)
-    if len(data)!=len(numpy.unique(hps,axis=0)):
-        print('WARNING: Some hyperparameter sets were evaluated more than once in %s (%i/%i)' % (path,len(numpy.unique(hps,axis=0)),len(data)))
+            if 'Test Loss' in line:
+                loss = float(line.split()[-1])
+                time = float(content[i+1].split()[-2])
+                data.append([loss,trial]+network_prms+samples+[nevals,time])
     if len(data)>0:
-        data = pd.DataFrame(data=data,columns=['params','loss','stdev']+['hp%i'%(i+1) for i in range(len(samples))]+['nevals','gen_params','dis_params']).sort_values(by=['nevals'])
+        data = pd.DataFrame(data=data,columns=['loss','trial']+network_names+['hp%i'%(i+1) for i in range(len(samples))]+['nevals','time']).sort_values(by=['nevals'])
     if debug:
         print(data)
     return data
@@ -179,6 +234,7 @@ def gather_trials(target,output,log_dir,ntasks,step,trial,split,**kwargs):
     return losses
 
 def extract_evals(log_dir,**kwargs):
+    uq = False
     samples, fvals, uqvals = [], [], []
     for file_name in sorted(glob.glob('%s/*.log' % log_dir)):
         file_content = numpy.loadtxt(file_name,dtype=str,delimiter='\n').tolist()
@@ -186,15 +242,16 @@ def extract_evals(log_dir,**kwargs):
             if 'Samples:' in line:
                 sample_set = line.split('Samples:')[-1].strip()[1:-1]
                 sample_set = [int(k) for k in sample_set.split()]
+            if 'UNCERTAINTY QUANTIFICATION' in line:
+                uq = True
             if 'Outer Loss' in line:
                 loss = float(line.split()[-1])
                 if math.isnan(loss)==False:
                     fvals.append(loss)
                     samples.append(sample_set)
-                    if 'Lower Bound of CI' in file_content[i+1]:
-                        lower = float(file_content[i+1].split()[-1])
-                        upper = float(file_content[i+2].split()[-1])
-                        uqvals.append([lower,upper])
+                    if uq:
+                        unc = float(file_content[i+2].split()[-1])
+                        uqvals.append([loss-2*unc,loss+2*unc])
     samples = numpy.array(samples,dtype=int)
     assert len(samples)>0, 'No samples found for surrogate modeling. Abort.'
     fvals = numpy.array(fvals).reshape(-1,1)
@@ -263,6 +320,24 @@ def get_samples(log_path,surrogate=False,mult=False):
         all_samples['sgate'] = surrogate[idxs,1:]
     return all_samples
 
+def data_and_model(path_to_config,samples,**kwargs):
+    with open(path_to_config) as f:
+        config = yaml.load(f, Loader=yaml.FullLoader)
+    data = get_data(**config['model'],**config['data'])
+    samples = ['%i'%samples['hp%i'%(i+1)] for i in range(len(config['prms']['names']))]
+    x_sc = {name:mult*int(sample) for name,mult,sample in zip(config['prms']['names'],config['prms']['mult'],samples)}
+    if 'default' in config['prms'].keys() and type(config['prms']['default'])==dict:
+        x_sc = {**x_sc,**config['prms']['default']}
+    hyperprms = get_hyperprms(config['trainer'], x_sc, **config['model'])
+    loaders = get_loader(data, **hyperprms, **config['model'])
+    preds = []
+    for i in range(config['model']['trial']):
+        out = 'evaluation_%s/%02i' % ('_'.join(samples),i+1)
+        y_real, y_pred = inference(data=loaders,hyperprms=hyperprms,log_dir=os.path.dirname(path_to_config)+'/logs',output=out,**config['model'])
+        y_real = y_real.detach().numpy().squeeze()
+        preds.append(y_pred.detach().numpy().squeeze())
+    return y_real, numpy.array(preds)
+
 def make_tables(table_type='evaluation'):
     os.makedirs('tables',exist_ok=True)
     if table_type == 'evaluation':
