@@ -8,10 +8,10 @@ import sys
 import logging
 import logging.config
 
-def create_logging(operation,log_dir='logs',step=0,nsteps=1,node_type='cpu',rank=0,size=1,**kwargs):
+def create_logging(operation,verbose=False,log_dir='logs',step=0,nsteps=1,node_type='cpu',rank=0,size=1,**kwargs):
     os.makedirs(log_dir,exist_ok=True)
     log_file = './%s/%s_%03i_%02i.log' % (log_dir,operation,step+1,rank+1)
-    config_logging(verbose=False,log_file=log_file)
+    config_logging(verbose=verbose,log_file=log_file)
     # Print out parallelization setup in log file
     logging.info('='*40)
     logging.info('DISTRIBUTION:')
